import json
from .jsonExtension import ExtensionBase

class ListExtension(ExtensionBase, list):
    def __new__(cls, other=None, saver=None):
        if isinstance(other, cls):
            return other
        return super().__new__(cls)

    def find(self, lmbd, *args, **kwargs):
        for item in self:
            if lmbd(item, *args, **kwargs):
                return item

    def join(self, separator: str, prefix="", postfix=""):
        string = ""
        for iterable, item in enumerate(self):
            string = f"{string}{prefix}{item}{postfix}"
            if iterable + 1 != len(self):
                string += separator
        return string

    def __call__(self):
        return ListExtension()

    def findall(self, lmbd, *args, **kwargs):
        lst = self()
        for item in self:
            if lmbd(item, *args, **kwargs):
                lst.append(item)
        return lst

    @classmethod
    def indexList(cls, list=None):
        iterate = list or cls
        l = cls()
        for index in range(len(iterate)):
            l.append(index)
        return l

    def all(self, function, *args, **kwargs):
        """
        The all function takes a function and it's additional argumes. 
        It returns True if the function(element, *args, **kwargs) evaluates to True for every element of the iterable, False otherwise.
        
        :param self: Used to Call the function on each element in the list.
        :param function: Used to Specify the function that is to be called on each element of the iterable.
        :param *args: Used to Pass a non-keyworded, variable-length argument list.
        :param **kwargs: Used to Pass keyworded, variable-length argument lists to the function.
        """
        for i in self:
            if not function(i, *args, **kwargs):
                return False
        return True

    def has(self, item, returnIndex=False):
        for i, iterator in enumerate(self):
            if hasattr(iterator, "has") and callable(iterator.has):
                if iterator.has(item):
                    return True if not returnIndex else i
            if iterator == item:
                return True if not returnIndex else i
        return False if not returnIndex else -1

    def indexOf(self, item):
        return self.has(item, True)

    def first(self):
        return self.get(0)

    def __getitem__(self, key):
        if isinstance(key, slice) or key < len(self):
            return list.__getitem__(self, key)
        return None

    get = __getitem__

    def filter(self, lmbd):
        instance = ListExtension()
        for i in self:
            if lmbd(i):
                instance.append(i)
        return instance

    @staticmethod
    def accept(data):
        return json.loads(data)

    def forEach(self, lmbd):
        for item in self:
            lmbd(item)

    def includes(self, value):
        return value in self

    @classmethod
    def byList(cls, lst):
        return cls(lst)

    def append(self, value):
        super().append(value)
        self.save()
        return self

    def copy(self):
        return ListExtension(self)

    def map(self, lmbd, *args):
        save = self()
        for element in self:
            save.append(lmbd(element, *args))
        return save

    def __add__(self, other):
        if type(other) is list:
            self += other
            self.save()
        else:
            self.append(other)
        return self



import json
import os


class ExtensionBase(object):

    classes = {}

    def __init_subclass__(cls) -> None:

        for cl in cls.mro():
            if cl not in (cls, ExtensionBase):
                cls.classes[cls] = cl
                cls._subclassing = cl
                break

    def __init__(self, iterable=None, saver=None):
        super().__init__(iterable) if iterable is not None else super().__init__()
        self.saver = saver

    def __getitem__(self, item):
        result = self.get(item)
        klass = StructByAction.get_class(result, init=False, saver=self.saver)
        if not isinstance(result, tuple(ExtensionBase.classes)):
            klass = StructByAction.get_class(result, saver=self.saver)
            self.setitem(item, klass, False)
            return klass
        return result

    @classmethod
    def accept(cls, data, **init_kwargs):
        for cls in cls.classes:
            if isinstance(data, cls):
                return data
            elif (clsm := getattr(cls, "accept", None)) is not None:
                try:
                    return StructByAction(clsm(data), **init_kwargs)
                except Exception:
                    continue
        return data

    def __setitem__(self, item, value, save=True):
        super().__setitem__(item, value)
        if save:
            self.save()

    def __delitem__(self, item):
        super().__delitem__(item)
        self.save()

    setitem = __setitem__

    def save(self):
        if self.saver is not None:
            self.saver.save()


class DictExtension(ExtensionBase, dict):
    @staticmethod
    def accept(data):
        return json.loads(data)


class StructByAction(object):

    __slots__ = "saver", "use_class"

    def __new__(cls, initData, saver=None):
        if type(initData) == type(cls.get_class(initData, init=False)):
            raise ValueError
        self_ins = super().__new__(cls)
        self_ins.saver = saver
        self_ins.use_class = cls.get_class(initData, saver=self_ins)
        return self_ins.use_class

    @staticmethod
    def get_class(data, *args, init=True, **kwargs):
        for klass in ExtensionBase.classes:
            if isinstance(data, klass._subclassing):
                return klass(data, *args, **kwargs) if init else klass
        return data

    def save(self):
        if self.saver is not None:
            self.saver(self.use_class)


def save(file, obj, indent=None):
    with open(file, "w", encoding="utf-8") as f:
        json.dump(obj, f, indent=indent)


def load(file, indent=None):
    with open(file, encoding="utf-8") as f:
        return StructByAction(json.load(f), saver=lambda d: save(file, d, indent))


def loadAdvanced(file, ident=None, content=None):
    if content is not None and not os.path.exists(file):
        os.makedirs(os.path.dirname(file), exist_ok=True)
        with open(file, "w", encoding="utf-8") as f:
            content = json.dumps(content) if isinstance(
                content, dict) else content
            f.write(content)
    return load(file, ident)
